#include <sys/stat.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
extern char *optarg;
extern int optind,opterr,optopt;
int getopt(int argc, char * const *argv, const char *optstring);

#include "cpmfs.h"

const char cmd[]="cpmcp";

/* containsDrive -- does this match a[0[0]]: ? */ /*{{{C}}}*//*{{{*/
static int containsDrive(const char *s)
{
  return
  (
    isalpha(*s) 
    && 
    (
      *(s+1)==':'
      || (isdigit(*(s+1)) && *(s+2)==':')
      || (isdigit(*(s+1)) && isdigit(*(s+2)) && *(s+3)==':')
    )
  );
}
/*}}}*/

int main(int argc, char *argv[])
{
  /* variables */ /*{{{*/
  int type;
  char *image;
  const char *format=FORMAT;
  int c,usage=0,readcpm=-1,todir=-1,text=0;
  struct cpmInode root;
  struct cpmSuperBlock drive;
  int exitcode=0;
  /*}}}*/

  /* parse options */ /*{{{*/
  type=CPMFS_DR22;
  while ((c=getopt(argc,argv,"pf:h?t"))!=EOF) switch(c)
  {
    case 'p': type=CPMFS_P2DOS; break;
    case 'f': format=optarg; break;
    case 'h':
    case '?': usage=1; break;
    case 't': text=1; break;
  }
  /*}}}*/
  /* parse arguments */ /*{{{*/
  if ((optind+2)>=argc) usage=1;
  else image=argv[optind];

  if (!usage)
  {
    if (containsDrive(argv[optind+1])) /* cpm -> unix */ /*{{{*/
    {
      int i;
      struct stat statbuf;

      for (i=optind+1; i<(argc-1); ++i) if (!containsDrive(argv[i])) { usage=1; break; }
      if (!usage)
      {
        todir=((argc-optind-1)>2);
        if (stat(argv[argc-1],&statbuf)==-1) { if (todir) usage=1; }
        else if (S_ISDIR(statbuf.st_mode)) todir=1; else if (todir) usage=1;
      }
      readcpm=1;
    }
    /*}}}*/
    else if (containsDrive(argv[argc-1])) /* unix -> cpm */ /*{{{*/
    {
      int i;

      todir=0;
      for (i=optind+1; i<(argc-1); ++i) if (containsDrive(argv[i])) { usage=1; break; }
      if (!usage && (argc-optind-1)>2 && *(strchr(argv[argc-1],':')+1)!='\0') usage=1;
      if (*(strchr(argv[argc-1],':')+1)=='\0') todir=1;
      readcpm=0;
    }
    /*}}}*/
    else usage=1;
  }

  if (usage)
  {
    fprintf(stderr,"Usage: %s [-f format] [-p] [-t] image a0:file file\n",cmd);
    fprintf(stderr,"       %s [-f format] [-p] [-t] image a0:file ... directory\n",cmd);
    fprintf(stderr,"       %s [-f format] [-p] [-t] image file a0:file\n",cmd);
    fprintf(stderr,"       %s [-f format] [-p] [-t] image file ... a0:\n",cmd);
    exit(1);
  }
  /*}}}*/
  /* open image file */ /*{{{*/
  if ((drive.fd = open(image,readcpm ? O_RDONLY : O_RDWR)) == -1 ) 
  {
    fprintf(stderr,"%s: can not open %s\n",cmd,image);
    exit(1);
  }
  getformat(format,type,&drive,&root);
  /*}}}*/
  if (readcpm) /* copy from CP/M to UNIX */ /*{{{*/
  {
    int i;

    for (i=optind+1; i<(argc-1); ++i)
    {
      char dest[_POSIX_PATH_MAX];
      struct cpmInode ino;

      if (todir) { strcpy(dest,argv[argc-1]); strcat(dest,"/"); }
      if (cpmNamei(&root,argv[i],&ino)==-1) { fprintf(stderr,"%s: can not open %s: %s\n",cmd,argv[i],boo); exitcode=1; }
      else
      {
        struct cpmFile file;
        FILE *ufp;

        cpmOpen(&ino,&file,O_RDONLY);
        if (todir) strcat(dest,argv[i]+2);
        else strcpy(dest,argv[argc-1]);
        if ((ufp=fopen(dest,text ? "w" : "wb"))==(FILE*)0) { fprintf(stderr,"%s: can not create %s: %s\n",cmd,dest,strerror(errno)); exitcode=1; }
        else
        {
          int crpending=0;
          int ohno=0;
          int res;
          char buf[4096];

          while ((res=cpmRead(&file,buf,sizeof(buf)))!=0)
          {
            int j;

            for (j=0; j<res; ++j)
            {
            if (text)
            {
              if (buf[j]=='\032') goto endwhile;
              if (crpending)
              {
                if (buf[j]=='\n') 
                {
                  if (putc('\n',ufp)==EOF) { fprintf(stderr,"%s: can not write %s: %s\n",cmd,dest,strerror(errno)); exitcode=1; ohno=1; goto endwhile; }
                  crpending=0;
                }
                else if (putc('\r',ufp)==EOF) { fprintf(stderr,"%s: can not write %s: %s\n",cmd,dest,strerror(errno)); exitcode=1; ohno=1; goto endwhile; }
                crpending=(buf[j]=='\r');
              }
              else
              {
                if (buf[j]=='\r') crpending=1;
                else if (putc(buf[j],ufp)==EOF) { fprintf(stderr,"%s: can not write %s: %s\n",cmd,dest,strerror(errno)); exitcode=1; ohno=1; goto endwhile; }
              }
            }
            else if (putc(buf[j],ufp)==EOF) { fprintf(stderr,"%s: can not write %s: %s\n",cmd,dest,strerror(errno)); exitcode=1; ohno=1; goto endwhile; }
            }
          }
          endwhile:
          if (fclose(ufp)==EOF && !ohno) { fprintf(stderr,"%s: can not close %s: %s\n",cmd,dest,strerror(errno)); exitcode=1; }
        }
        cpmClose(&file);
      }
    }
  }
  /*}}}*/
  else /* copy from UNIX to CP/M */ /*{{{*/
  {
    int i;

    for (i=optind+1; i<(argc-1); ++i)
    {
      /* variables */ /*{{{*/
      char *dest;
      FILE *ufp;
      /*}}}*/

      if ((ufp=fopen(argv[i],"r"))==(FILE*)0) /* cry a little */ /*{{{*/
      {
        fprintf(stderr,"%s: can not open %s: %s\n",cmd,argv[i],strerror(errno));
        exitcode=1;
      }
      /*}}}*/
      else
      {
        struct cpmInode ino;
        char cpmname[1+2+1+8+3+1]; /* a31:foobarxy.zzy\0 */

        if (todir)
        {
          if ((dest=strrchr(argv[i],'/'))!=(char*)0) ++dest; else dest=argv[i];
          strcpy(cpmname,argv[argc-1]);
          strcat(cpmname,dest);
          dest=cpmname;
        }
        else dest=argv[argc-1];
        if (cpmCreat(&root,dest,&ino,0666)==-1) /* just cry */ /*{{{*/
        {
          fprintf(stderr,"%s: can not create %s: %s\n",cmd,dest,boo);
          exitcode=1;
        }
        /*}}}*/
        else
        {
          struct cpmFile file;
          int ohno=0;
          char buf[4096+1];

          cpmOpen(&ino,&file,O_WRONLY);
          do
          {
          int j;

          for (j=0; j<(sizeof(buf)/2) && (c=getc(ufp))!=EOF; ++j)
          {
            if (text && c=='\n') buf[j++]='\r';
            buf[j]=c;
          }
          if (text && c==EOF) buf[j++]='\032';
          if (cpmWrite(&file,buf,j)!=j)
          {
            fprintf(stderr,"%s: can not write %s: %s\n",cmd,dest,boo);
            ohno=1;
            exitcode=1;
            break;
          }
          } while (c!=EOF);
          if (cpmClose(&file)==EOF && !ohno) /* I just can't hold back the tears */ /*{{{*/
          {
            fprintf(stderr,"%s: can not close %s: %s\n",cmd,dest,boo);
            exitcode=1;
          }
          /*}}}*/
        }
        fclose(ufp);
      }
    }
  }
  /*}}}*/
  exit(exitcode);
}
