/* #includes */ /*{{{C}}}*//*{{{*/
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
extern char *optarg;
extern int optind,opterr,optopt;
int getopt(int argc, char * const *argv, const char *optstring);

#include "cpmfs.h"
/*}}}*/

/* variables */ /*{{{*/
static const char * const month[12]={"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec" };
/*}}}*/

/* namecmp -- compare two entries */ /*{{{*/
static int namecmp(const void *a, const void *b)
{
  return (strcmp(((const struct cpmDirent*)a)->name,((const struct cpmDirent*)b)->name));
}
/*}}}*/
/* olddir  -- old style output */ /*{{{*/
static void olddir(struct cpmDirent *dirent, int entries)
{
  int i,j,k,l;

  for (i=l=0; i<entries; ++i)
  {
    if (l%4) printf(" : ");
    if (dirent[i].name[0]!='.')
    {
    for (j=0; dirent[i].name[j] && dirent[i].name[j]!='.'; ++j) putchar(toupper(dirent[i].name[j]));
    k=j; while (k<9) { putchar(' '); ++k; }
    if (dirent[i].name[j]=='.') ++j;
    for (k=0; dirent[i].name[j]; ++j,++k) putchar(toupper(dirent[i].name[j]));
    for (; k<3; ++k) putchar(' ');
    ++l;
    }
    if (!((l+1)%4)) putchar('\n');
  }
  if (l%4) putchar('\n');
  if (entries==0) printf("No files\n");
}
/*}}}*/
/* oldddir -- old style long output */ /*{{{*/
static void oldddir(struct cpmDirent *dirent, int entries, struct cpmInode *ino)
{
  struct cpmStatFS buf;
  struct cpmStat statbuf;
  struct cpmInode file;

  if (entries)
  {
  int i,j,k,l;

  qsort(dirent,entries,sizeof(struct cpmDirent),namecmp);
  printf("     Name    Bytes   Recs  Attr User       Update             Create\n");
  printf("------------ ------ ------ ---- ----  -----------------  -----------------\n");
  cpmStatFS(ino,&buf);
  for (i=l=0; i<entries; ++i)
  {
    struct tm *tmp;

if (dirent[i].name[0]!='.')
{
    for (j=0; dirent[i].name[j] && dirent[i].name[j]!='.'; ++j) putchar(toupper(dirent[i].name[j]));
    k=j; while (k<8) { putchar(' '); ++k; }
    putchar('.');
    if (dirent[i].name[j]=='.') ++j;
    for (k=0; dirent[i].name[j]; ++j,++k) putchar(toupper(dirent[i].name[j]));
    for (; k<3; ++k) putchar(' ');

    cpmNamei(ino,dirent[i].name,&file);
    cpmStat(&file,&statbuf);
    printf(" %5.1ldK",(statbuf.size+buf.f_bsize-1)/buf.f_bsize*(buf.f_bsize/1024));
    printf(" %6.1ld ",(long)(statbuf.size/128));
    putchar(statbuf.mode&0200 ? ' ' : 'R');
    putchar(' ');
    putchar(' ');
    printf("  %4.1d",0);
    tmp=gmtime(&statbuf.mtime);
    printf("  %02d-%s-%04d %02d:%02d",tmp->tm_mday,month[tmp->tm_mon],tmp->tm_year+1900,tmp->tm_hour,tmp->tm_min);
    tmp=gmtime(&statbuf.ctime);
    printf("  %02d-%s-%04d %02d:%02d",tmp->tm_mday,month[tmp->tm_mon],tmp->tm_year+1900,tmp->tm_hour,tmp->tm_min);
    putchar('\n');
++l;
}
  }
  printf("%5.1d Files occupying %6.1ldK",l,(buf.f_bused*buf.f_bsize)/1024);
  printf(", %7.1ldK Free.\n",(buf.f_bfree*buf.f_bsize)/1024);
}
else printf("No files found\n");
}
/*}}}*/
/* ls      -- UNIX style output */ /*{{{*/
static void ls(struct cpmDirent *dirent, int entries, struct cpmInode *ino, int l, int c)
{
  int i;
  time_t now;
  struct cpmStat statbuf;
  struct cpmInode file;

  time(&now);
  qsort(dirent,entries,sizeof(struct cpmDirent),namecmp);
  for (i=0; i<entries; ++i) if (dirent[i].name[0]!='.')
  {
    if (l)
    {
      struct tm *tmp;

      cpmNamei(ino,dirent[i].name,&file);
      cpmStat(&file,&statbuf);
      putchar(S_ISDIR(statbuf.mode) ? 'd' : '-');
      putchar(statbuf.mode&0400 ? 'r' : '-');
      putchar(statbuf.mode&0200 ? 'w' : '-');
      putchar(statbuf.mode&0100 ? 'x' : '-');
      putchar(statbuf.mode&0040 ? 'r' : '-');
      putchar(statbuf.mode&0020 ? 'w' : '-');
      putchar(statbuf.mode&0010 ? 'x' : '-');
      putchar(statbuf.mode&0004 ? 'r' : '-');
      putchar(statbuf.mode&0002 ? 'w' : '-');
      putchar(statbuf.mode&0001 ? 'x' : '-');
#if 0
      putchar(statbuf.flags&FLAG_PUBLIC ? 'p' : '-');
      putchar(dir[i].flags&FLAG_SYSTEM ? 's' : '-');
      printf(" %-2d ",dir[i].user);
#endif
      printf("%8.1ld ",(long)statbuf.size);
      tmp=gmtime(c ? &statbuf.ctime : &statbuf.mtime);
      printf("%s %02d ",month[tmp->tm_mon],tmp->tm_mday);
      if ((c ? statbuf.ctime : statbuf.mtime)<(now-182*24*3600)) printf("%04d  ",tmp->tm_year+1900);
      else printf("%02d:%02d ",tmp->tm_hour,tmp->tm_min);
    }
    printf("%s\n",dirent[i].name);
  }
}
/*}}}*/

const char cmd[]="cpmls";

int main(int argc, char *argv[])
{
  /* variables */ /*{{{*/
  char *image;
  const char *format=FORMAT;
  int c,usage=0;
  struct cpmSuperBlock drive;
  struct cpmInode root;
  int entries,dirsize=0;
  int style=0;
  int changetime=0;
  struct cpmFile dir;
  struct cpmDirent *dirent=(struct cpmDirent*)0;
  /*}}}*/

  /* parse options */ /*{{{*/
  while ((c=getopt(argc,argv,"cf:h?oOl"))!=EOF) switch(c)
  {
    case 'f': format=optarg; break;
    case 'h':
    case '?': usage=1; break;
    case 'o': style=1; break;
    case 'O': style=2; break;
    case 'l': style=3; break;
    case 'c': changetime=1;
  }

  if (optind!=(argc-1)) usage=1;
  else image=argv[optind];

  if (usage)
  {
    fprintf(stderr,"Usage: %s [-f format] [-o|-O|-l [-c]] image\n",cmd);
    exit(1);
  }
  /*}}}*/
  /* open image */ /*{{{*/
  if ((drive.fd = open(image,O_RDONLY)) == -1 ) 
  {
    fprintf(stderr,"%s: can not open %s: %s\n",cmd,image,strerror(errno));
    exit(1);
  }
  getformat(format,CPMFS_P2DOS,&drive,&root);
  /*}}}*/
  cpmOpendir(&root,&dir);
  entries=0;
  dirent=realloc(dirent,sizeof(struct cpmDirent)*(dirsize=32));
  while (cpmReaddir(&dir,dirent+entries))
  {
    ++entries;
    if (entries==dirsize) dirent=realloc(dirent,sizeof(struct cpmDirent)*(dirsize*=2));
  }
  if (style==1) olddir(dirent,entries);
  else if (style==2) oldddir(dirent,entries,&root);
  else ls(dirent,entries,&root,style==3,changetime);
  exit(0);
}
