/* #includes */ /*{{{C}}}*//*{{{*/
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
extern char *optarg;
extern int optind,opterr,optopt;
int getopt(int argc, char * const *argv, const char *optstring);

#include "cpmfs.h"
/*}}}*/

/* mkfs -- make file system */ /*{{{*/
static int mkfs(struct cpmSuperBlock *drive, const char *name, int type)
{
  /* variables */ /*{{{*/
  int i;
  char buf[128];
  int fd;
  int bytes;
  int trkbytes;
  /*}}}*/

  /* open image file */ /*{{{*/
  if ((fd = open(name, O_CREAT|O_WRONLY, 0666)) < 0)
  {
    boo=strerror(errno);
    return -1;
  }
  /*}}}*/
  /* write system tracks */ /*{{{*/
  memset(buf,0xe5,128);
  /* this initialises only whole tracks, so it skew is not an issue */
  trkbytes=drive->secLength*drive->sectrk;
  for (i=0; i < trkbytes*drive->boottrk; i += 128) if (write(fd, buf, 128)!=128)
  {
    boo=strerror(errno);
    close(fd);
    return -1;
  }
  /*}}}*/
  /* write direcctory */ /*{{{*/
  bytes=drive->maxdir*32;
  if (bytes%trkbytes) bytes=((bytes+trkbytes)/trkbytes)*trkbytes;
  if (type==CPMFS_P2DOS) buf[3*32]=0x21;
  for (i=0; i < bytes; i += 128) if (write(fd, buf, 128)!=128)
  {
    boo=strerror(errno);
    close(fd);
    return -1;
  }
  /*}}}*/
  /* close image file */ /*{{{*/
  if (close(fd)==-1)
  {
    boo=strerror(errno);
    return -1;
  }
  /*}}}*/
  return 0;
}
/*}}}*/

const char cmd[]="mkfs.cpm";

int main(int argc, char *argv[])
{
  char *image;
  const char *format=FORMAT;
  int c,usage=0;
  struct cpmSuperBlock drive;
  struct cpmInode root;
  int type=CPMFS_DR22;

  while ((c=getopt(argc,argv,"pf:h?"))!=EOF) switch(c)
  {
    case 'p': type=CPMFS_P2DOS; break;
    case 'f': format=optarg; break;
    case 'h':
    case '?': usage=1; break;
  }

  if (optind!=(argc-1)) usage=1;
  else image=argv[optind];

  if (usage)
  {
    fprintf(stderr,"Usage: %s [-f format] [-p] image\n",cmd);
    exit(1);
  }
  drive.fd=-1;
  getformat(format,type,&drive,&root);
  if (mkfs(&drive,image,type)==-1)
  {
    fprintf(stderr,"%s: can not make new file system: %s\n",cmd,boo);
    exit(1);
  }
  else exit(0);
}
